/*
 * This file is part of the Cardinal Optimizer, all rights reserved.
 */

// We will compute feasibility relaxation for the following infeasible LP problem,
// the problem is 'itest6' test case from netlib-infeas:
//
// Minimize
//  X2 + X3 + X4
// Subject To
//  ROW1: 0.8 X3 + X4 <= 10000
//  ROW2: X1 <= 90000
//  ROW3: 2 X6 - X8 <= 10000
//  ROW4: - X2 + X3 >= 50000
//  ROW5: - X2 + X4 >= 87000
//  ROW6: X3 <= 50000
//  ROW7: - 3 X5 + X7 >= 10000
//  ROW8: 0.5 X5 + 0.6 X6 <= 300000
//  ROW9: X2 - 0.05 X3 = 5000
//  ROW10: X2 - 0.04 X3 - 0.05 X4 = 4500
//  ROW11: X2 >= 80000
// END

#include "coptcpp_pch.h"

using namespace std;

int main(int argc, char *argv[]) {
  try {
    Envr env;
    Model model = env.CreateModel("feasrelax_ex1");

    // Add variables
    Var x1 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X1");
    Var x2 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X2");
    Var x3 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X3");
    Var x4 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X4");
    Var x5 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X5");
    Var x6 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X6");
    Var x7 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X7");
    Var x8 = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "X8");

    // Add constraints
    model.AddConstr(0.8*x3 + x4 <= 10000, "ROW1");
    model.AddConstr(x1 <= 90000, "ROW2");
    model.AddConstr(2*x6 - x8 <= 10000, "ROW3");
    model.AddConstr(-x2 + x3 >= 50000, "ROW4");
    model.AddConstr(-x2 + x4 >= 87000, "ROW5");
    model.AddConstr(x3 <= 50000, "ROW6");
    model.AddConstr(-3*x5 + x7 >= 10000, "ROW7");
    model.AddConstr(0.5*x5 + 0.6*x6 <= 300000, "ROW8");
    model.AddConstr(x2 - 0.05*x3 == 5000, "ROW9");
    model.AddConstr(x2 - 0.04*x3 - 0.05*x4 == 4500, "ROW10");
    model.AddConstr(x2 >= 80000, "ROW11");

    // Set objective
    model.SetObjective(x2 + x3 + x4, COPT_MINIMIZE);

    // Set parameters
    model.SetDblParam(COPT_DBLPARAM_TIMELIMIT, 60);

    // Solve the problem
    model.Solve();

    // Compute feasibility relaxation if problem is infeasible
    if (model.GetIntAttr(COPT_INTATTR_LPSTATUS) == COPT_LPSTATUS_INFEASIBLE) {
      // Set feasibility relaxation mode: minimize sum of violations
      model.SetIntParam(COPT_INTPARAM_FEASRELAXMODE, 0);

      // Compute feasibility relaxation
      model.FeasRelax(1, 1);

      // Check if feasibility relaxation solution is available
      if (model.GetIntAttr(COPT_INTATTR_HASFEASRELAXSOL)) {
        ConstrArray cons = model.GetConstrs();
        VarArray vars = model.GetVars();

        // Retrieve feasibility relaxation result
        cout << endl;
        cout << "======================== FeasRelax result ========================" << endl;

        for (int iRow = 0; iRow < cons.Size(); ++iRow) {
          Constraint con = cons.GetConstr(iRow);
          double lowRelax = con.Get(COPT_DBLINFO_RELAXLB);
          double uppRelax = con.Get(COPT_DBLINFO_RELAXUB);
          if (lowRelax != 0.0 || uppRelax != 0.0) {
            cout << "  " << con.GetName() << ": " << "viol = " << "(" << lowRelax << ", " << uppRelax << ")" << endl;
          }
        }

        cout << endl;
        for (int iCol = 0; iCol < vars.Size(); ++iCol) {
          Var var = vars.GetVar(iCol);
          double lowRelax = var.Get(COPT_DBLINFO_RELAXLB);
          double uppRelax = var.Get(COPT_DBLINFO_RELAXUB);
          if (lowRelax != 0.0 || uppRelax != 0.0) {
            cout << "  " << var.GetName() << ": " << "viol = " << "(" << lowRelax << ", " << uppRelax << ")" << endl;
          }
        }

        // Write relaxed problem to file
        model.WriteRelax("feasrelax_ex1.relax");
      }
    }
  }
  catch (CoptException e) {
    cout << "Error Code = " << e.GetCode() << endl;
    cout << e.what() << endl;
  }
  catch (...) {
    cout << "Unknown exception occurs!";
  }
}
